/*
 * Copyright (C) 1997
 *	Sony Computer Science Laboratories Inc.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY SONY CSL AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL SONY CSL OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: altq_wfq.c,v 1.4 1998/09/23 07:56:40 kjc Exp kjc $
 */
/*
 *  March 27, 1997.  Written by Hiroshi Kyusojin of Keio University
 *  (kyu@mt.cs.keio.ac.jp). 
 */

#ifdef __FreeBSD__
#include "opt_altq.h"
#endif
#ifdef WFQ

#include <sys/types.h>
#include <sys/param.h>
#include <sys/malloc.h>
#include <sys/mbuf.h>
#include <sys/uio.h>
#include <sys/socket.h>
#include <sys/systm.h>
#include <sys/proc.h>
#include <sys/errno.h>
#include <sys/time.h>
#include <sys/kernel.h>

#include <net/if.h>
#include <net/if_types.h>
#include <netinet/altq_wfq.h>
/*
#define WFQ_DEBUG
*/

static int		wfq_setenable(struct wfq_interface *, int);
static int		wfq_ifattach(struct wfq_interface *);
static int		wfq_ifdetach(struct wfq_interface *);
static int		wfq_ifenqueue(struct ifnet *, struct mbuf *, struct pr_hdr *, int);
u_long			wfq_hash(struct flowinfo *, int);
u_long			wfq_hashbydstaddr(struct flowinfo *, int);
u_long			wfq_hashbysrcport(struct flowinfo *, int);
static wfq		*wfq_maxqueue(wfq_state_t *);
static struct mbuf	*wfq_ifdequeue(struct ifnet *, int);
static int		wfq_getqid(struct wfq_getqid *);
static int		wfq_setweight(struct wfq_setweight *);
static int		wfq_getstats(struct wfq_getstats *);
static int		wfq_changehashfnc(struct wfq_changehashfnc *);
static int		wfq_flush(struct ifnet *);

/* global value : pointer to wfq queue list */
static wfq_state_t *wfq_list = NULL;


static int
wfq_setenable(ifacep, flag)
	struct wfq_interface *ifacep;
	int flag;
{
	struct ifnet *ifp;
	wfq_state_t *wfqp;
	int error = 0;

	if ((ifp = ifunit(ifacep->wfq_ifacename)) == NULL)
		return (ENXIO);

	wfqp = (wfq_state_t *)ifp->if_altqp;

	switch(flag){
	case ENABLE:
		error = if_altqenable(wfqp->ifp);
		break;
	case DISABLE:
		error = if_altqdisable(wfqp->ifp);
		break;
	}
	return error;
}


static int
wfq_ifattach(ifacep)
	struct wfq_interface *ifacep;
{
	int error = 0, i;
	struct ifnet *ifp;
	wfq_state_t *new_wfqp;
	wfq *queue;

	if ((ifp = ifunit(ifacep->wfq_ifacename)) == NULL) {
#ifdef WFQ_DEBUG
		printf("wfq_ifattach()...no ifp found\n");
#endif
		return (ENXIO);
	}

	if (!ALTQ_IS_READY(ifp)) {
#ifdef WFQ_DEBUG
		printf("wfq_ifattach()...altq is not ready\n");
#endif
		return (ENXIO);
	}

	if(ifacep->nqueues <= 0 ||  MAX_QSIZE < ifacep->nqueues)
		ifacep->nqueues = DEFAULT_QSIZE;

	/* allocate and initialize wfq_state_t */
	MALLOC(new_wfqp, wfq_state_t *, sizeof(wfq_state_t),
	       M_DEVBUF, M_WAITOK);
	bzero(new_wfqp, sizeof(wfq_state_t));
	MALLOC(queue, wfq *, sizeof(wfq) * ifacep->nqueues,
	       M_DEVBUF, M_WAITOK);
	bzero(queue, sizeof(wfq) * ifacep->nqueues);

	/* keep the ifp */
	new_wfqp->ifp = ifp;
	new_wfqp->nums = ifacep->nqueues;
	new_wfqp->hwm = HWM;
	new_wfqp->bytes = 0;
	new_wfqp->rrp = NULL;
	new_wfqp->queue = queue;
	new_wfqp->hash_func = wfq_hashbydstaddr;

	for (i = 0; i < new_wfqp->nums; i++, queue++) {
		queue->next = queue->prev = NULL;
		queue->head = queue->tail = NULL;
		queue->bytes = queue->quota = 0;
		queue->weight = 100;
	}

	/*
	 * set WFQ to this ifnet structure.
	 */
	if ((error = if_altqattach(ifp, new_wfqp, wfq_ifenqueue, wfq_ifdequeue,
				  ALTQT_WFQ)) != 0) {
		FREE(queue, M_DEVBUF);
		FREE(new_wfqp, M_DEVBUF);
		return (error);
	}

	new_wfqp->next = wfq_list;
	wfq_list = new_wfqp;

	return (error);
}


static int
wfq_ifdetach(ifacep)
	struct wfq_interface *ifacep;
{
	int		error = 0;
	wfq_state_t	*wfqp;
	struct ifnet	*ifp;

	if ((ifp = ifunit(ifacep->wfq_ifacename)) == NULL)
		return (ENXIO);

	wfqp = (wfq_state_t *)ifp->if_altqp;

	/* free queued mbuf */
	wfq_flush(ifp);

	/* remove WFQ from the ifnet structure. */
	(void)if_altqdisable(ifp);
	(void)if_altqdetach(ifp);
    
	/* remove from the wfqstate list */
	if (wfq_list == wfqp)
		wfq_list = wfqp->next;
	else {
		wfq_state_t *wp = wfq_list;
		do {
			if (wp->next == wfqp) {
				wp->next = wfqp->next;
				break;
			}
		} while ((wp = wp->next) != NULL);
	}
    
	/* deallocate wfq_state_t */
	FREE(wfqp->queue, M_DEVBUF);
	FREE(wfqp, M_DEVBUF);
	return (error);
}

static int
wfq_flush(ifp)
	struct ifnet *ifp;
{
	struct mbuf *mp;

	while ((mp = wfq_ifdequeue(ifp, ALTDQ_DEQUEUE)) != NULL)
		m_freem(mp);
	return 0;
}

static int
wfq_ifenqueue(ifp, mp, pr_hdr, mode)
	struct ifnet *ifp;
	struct mbuf *mp;
	struct pr_hdr *pr_hdr;
	int mode;
{
	wfq_state_t *wfqp;
	struct flowinfo flow;
	wfq *queue;
	int byte, error = 0;

	if (mode != ALTEQ_NORMAL)
		return 0;

	wfqp = (wfq_state_t *)ifp->if_altqp;
	mp->m_nextpkt = NULL;

#define WFQ_FMASK (FIMB4_PROTO|FIMB4_SADDR|FIMB4_DADDR|FIMB4_SPORT|FIMB4_DPORT)
	altq_extractflow(mp, pr_hdr, &flow, WFQ_FMASK);
	queue = &wfqp->queue[(*wfqp->hash_func)(&flow, wfqp->nums)];
	if (queue->tail == NULL)
		queue->head = mp;
	else
		queue->tail->m_nextpkt = mp;
	queue->tail = mp;
	byte = mp->m_pkthdr.len;
	queue->bytes += byte;
	wfqp->bytes += byte;

	if (queue->next == NULL) {
		/* this queue gets active. add the queue to the active list */
		if (wfqp->rrp == NULL){
			/* no queue in the active list */
			queue->next = queue->prev = queue;
			wfqp->rrp = queue;
			WFQ_ADDQUOTA(queue);
		}
		else{
			/* insert the queue at the tail of the active list */
			queue->prev = wfqp->rrp->prev;
			wfqp->rrp->prev->next = queue;
			wfqp->rrp->prev = queue;
			queue->next = wfqp->rrp;
			queue->quota = 0;
		}
	}

	/* check overflow. if the total size exceeds the high water mark,
	   drop packets from the longest queue. */
	while (wfqp->bytes > wfqp->hwm) {
		wfq *drop_queue = wfq_maxqueue(wfqp);

		/* drop the packet at the head. */
		mp = drop_queue->head;
		if ((drop_queue->head = mp->m_nextpkt) == NULL)
			drop_queue->tail = NULL;
		mp->m_nextpkt = NULL;
		byte = mp->m_pkthdr.len;
		drop_queue->bytes -= byte;
		drop_queue->drop_packets++;
		drop_queue->drop_bytes += byte;
		wfqp->bytes -= byte;
		m_freem(mp);
		if(drop_queue == queue)
			/* the queue for this flow is selected to drop */
			error = ENOBUFS;
	}

	/*
	 * call the driver's start routine.
	 */
	ifp = wfqp->ifp;
	if(ifp->if_start && (ifp->if_flags & IFF_OACTIVE) == 0)
		(*ifp->if_start)(ifp);

	return error;
}
	

u_long wfq_hash(flow, n)
	struct flowinfo *flow;
	int n;
{
	struct flowinfo_in *fp;
	u_long val, val2;

	if (flow == NULL || flow->fi_family != AF_INET)
		/* currently only AF_INET is supported, use qid 0. */
		return 0;

	fp = (struct flowinfo_in *)flow;
	val = fp->fi_dst.s_addr ^ fp->fi_src.s_addr;
	val = val ^ (val >> 8) ^ (val >> 16) ^ (val >> 24);
	val2 = fp->fi_dport ^ fp->fi_sport ^ fp->fi_proto;
	val2 = val2 ^ (val2 >> 8);
	val = val ^ val2;
	return (val % n);
}


u_long wfq_hashbydstaddr(flow, n)
	struct flowinfo *flow;
	int n;
{
	struct flowinfo_in *fp;
	u_long val;

	if (flow == NULL || flow->fi_family != AF_INET)
		/* currently only AF_INET is supported, use qid 0. */
		return 0;

	fp = (struct flowinfo_in *)flow;
	val = fp->fi_dst.s_addr;
	val = val ^ (val >> 8) ^ (val >> 16) ^ (val >> 24);

	return (val % n);
}

u_long wfq_hashbysrcport(flow, n)
	struct flowinfo *flow;
	int n;
{
	struct flowinfo_in *fp;
	u_long val;

	if (flow == NULL || flow->fi_family != AF_INET)
		/* currently only AF_INET is supported, use qid 0. */
		return 0;

	fp = (struct flowinfo_in *)flow;
	val = fp->fi_sport;
	val = val ^ (val >> 8);

	return (val % n);
}

static wfq *wfq_maxqueue(wfqp)
	wfq_state_t *wfqp;
{
	int byte, max_byte = 0;
	wfq *queue, *max_queue = NULL;

	if((queue = wfqp->rrp) == NULL)
		/* never happens */
		return NULL;
	do{
		if ((byte = queue->bytes * 100 / queue->weight) > max_byte) {
			max_queue = queue;
			max_byte = byte;
		}
	} while ((queue = queue->next) != wfqp->rrp);

	return max_queue;
}


static struct mbuf *
wfq_ifdequeue(ifp, mode)
	struct ifnet *ifp;
	int mode;
{
	wfq_state_t *wfqp;
	wfq *queue;
	struct mbuf *mp;
	int byte;

	wfqp = (wfq_state_t *)ifp->if_altqp;

	if (mode == ALTDQ_FLUSH) {
		wfq_flush(ifp);
		return NULL;
	}

	if ((wfqp->bytes == 0) || ((queue = wfqp->rrp) == NULL))
		/* no packt in the queues */
		return NULL;
	
	while (1) {
		if(queue->quota > 0) {
			if(queue->bytes <= 0) {
				/* this queue no longer has packet.
				   remove the queue from the active list. */
				if(queue->next == queue){
					/* no other active queue 
					   -- this case never happens in
					   this algorithm. */
					queue->next = queue->prev = NULL;
					wfqp->rrp = NULL;
					return NULL;
				}
				else {
					queue->prev->next = queue->next;
					queue->next->prev = queue->prev;
					/* the round-robin pointer points
					   to this queue, advance the rrp */
					wfqp->rrp = queue->next;
					queue->next = queue->prev = NULL;
					queue = wfqp->rrp;
					WFQ_ADDQUOTA(queue);
					continue;
				}
			}

			/* dequeue a packet from this queue */
			mp = queue->head;
			if (mode == ALTDQ_DEQUEUE) {
				if((queue->head = mp->m_nextpkt) == NULL)
					queue->tail = NULL;
				byte = mp->m_pkthdr.len;
				mp->m_nextpkt = NULL;
				queue->quota -= byte;
				queue->bytes -= byte;
				queue->sent_packets++;
				queue->sent_bytes += byte;
				wfqp->bytes -= byte;
			}
			return mp;

			/* if the queue gets empty by this dequeueing,
			   the queue will be removed from the active list
			   at the next round */
		}
	
		/* advance the round-robin pointer */
		queue = wfqp->rrp = queue->next;
		WFQ_ADDQUOTA(queue);
	}
}

static int
wfq_getqid(gqidp)
	struct wfq_getqid *gqidp;
{
	struct ifnet *ifp;
	wfq_state_t *wfqp;

	if ((ifp = ifunit(gqidp->iface.wfq_ifacename)) == NULL)
		return (ENXIO);

	wfqp = (wfq_state_t *)ifp->if_altqp;
	gqidp->qid = (*wfqp->hash_func)(&gqidp->flow, wfqp->nums);
	return 0;
}

static int
wfq_setweight(swp)
	struct wfq_setweight *swp;
{
	struct ifnet *ifp;
	wfq *queue;
	int old;

	if (swp->weight < 0) {
		printf("set weight in natural number\n");
		return (ENXIO);
	}

	if ((ifp = ifunit(swp->iface.wfq_ifacename)) == NULL)
		return (ENXIO);

	queue = &((wfq_state_t *)ifp->if_altqp)->queue[swp->qid];
	old = queue->weight;
	queue->weight = swp->weight;
	swp->weight = old;
	return 0;
}


static int
wfq_getstats(gsp)
	struct wfq_getstats *gsp;
{
	struct ifnet *ifp;
	wfq *queue;
	queue_stats *stats;

	if ((ifp = ifunit(gsp->iface.wfq_ifacename)) == NULL)
		return (ENXIO);

	queue = &((wfq_state_t *)ifp->if_altqp)->queue[gsp->qid];
	stats = &gsp->stats;

	stats->bytes		= queue->bytes;
	stats->weight		= queue->weight;
	stats->sent_packets	= queue->sent_packets;
	stats->sent_bytes	= queue->sent_bytes;
	stats->drop_packets	= queue->drop_packets;
	stats->drop_bytes	= queue->drop_bytes;

	return 0;
}


static int
wfq_changehashfnc(chf)
	struct wfq_changehashfnc *chf;
{
	struct ifnet *ifp;
	int error = 0;

	if ((ifp = ifunit(chf->iface.wfq_ifacename)) == NULL)
		return (ENXIO);

	switch (chf->hash_policy) {
	case WFQ_HASH_DSTADDR:
		((wfq_state_t *)ifp->if_altqp)->hash_func = wfq_hashbydstaddr;
		break;
	case WFQ_HASH_SRCPORT:
		((wfq_state_t *)ifp->if_altqp)->hash_func = wfq_hashbysrcport;
		break;
	case WFQ_HASH_FULL:
		((wfq_state_t *)ifp->if_altqp)->hash_func = wfq_hash;
		break;
	default:
		error = EINVAL;
		break;
	}
	return error;
}



/*
 * wfq device interface
 */
#include <sys/ioccom.h>
#include <sys/conf.h>

d_open_t	wfqopen;
d_close_t	wfqclose;
d_ioctl_t	wfqioctl;

int
wfqopen(dev, flag, fmt, p)
	dev_t dev;
	int flag, fmt;
	struct proc *p;
{
	return 0;
}

int
wfqclose(dev, flag, fmt, p)
	dev_t dev;
	int flag, fmt;
	struct proc *p;
{
	struct ifnet *ifp;
	struct wfq_interface iface;
	wfq_state_t *wfqp;
	int s;
    
	s = splimp();
	while ((wfqp = wfq_list) != NULL) {
		ifp = wfqp->ifp;
#ifdef __NetBSD__
		sprintf(iface.wfq_ifacename, "%s", ifp->if_xname);
#else
		sprintf(iface.wfq_ifacename, "%s%d",
			ifp->if_name, ifp->if_unit);
#endif
		iface.wfq_ifacelen = strlen(iface.wfq_ifacename);
		wfq_ifdetach(&iface);
	}
	splx(s);
	return 0;
}

int
wfqioctl(dev, cmd, addr, flag, p)
	dev_t dev;
	ioctlcmd_t cmd;
	caddr_t addr;
	int flag;
	struct proc *p;
{
	int	error = 0;
	int 	s;

	/* check cmd for superuser only */
	switch (cmd) {
	case WFQ_GET_QID:
	case WFQ_GET_STATS:
		break;
	default:
		if ((error = suser(p->p_ucred, &p->p_acflag)) != 0) {
#ifdef WFQ_DEBUG
			printf("wfqioctl()...not super user\n");
#endif
			return (error);
		}
		break;
	}

	s = splimp();
	switch (cmd) {
		
	case WFQ_ENABLE:
		error = wfq_setenable((struct wfq_interface *)addr, ENABLE);
		break;

	case WFQ_DISABLE:
		error = wfq_setenable((struct wfq_interface *)addr, DISABLE);
		break;

	case WFQ_IF_ATTACH:
		error = wfq_ifattach((struct wfq_interface *)addr);
		break;

	case WFQ_IF_DETACH:
		error = wfq_ifdetach((struct wfq_interface *)addr);
		break;

	case WFQ_GET_QID:
		error = wfq_getqid((struct wfq_getqid *)addr);
		break;

	case WFQ_SET_WEIGHT:
		error = wfq_setweight((struct wfq_setweight *)addr);
		break;

	case WFQ_GET_STATS:
		error = wfq_getstats((struct wfq_getstats *)addr);
		break;

	case WFQ_CHANGE_HASHFNC:
		error = wfq_changehashfnc((struct wfq_changehashfnc *)addr);
		break;

	default:
		error = EINVAL;
		break;
	}
	splx(s);
	return error;
}

#endif /* WFQ */
