/*
 *  Intel PCIC or compatible Controller driver
 *  May be built using LKM to make a loadable module.
 *-------------------------------------------------------------------------
 *
 * Copyright (c) 1995 Andrew McRae.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * pcic98 : PC9801 original PCMCIA controller code for NS/A,Ne,NX/C,NR/L.
 * by Noriyuki Hosobuchi <yj8n-hsbc@asahi-net.or.jp>
 */

#include "pcic.h"
#if NPCIC > 0

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/kernel.h>
#include <sys/select.h>

#include <machine/clock.h>

#include <i386/isa/icu.h>
#include <i386/isa/isa_device.h>

#include <pccard/i82365reg.h>
#if 0
#ifdef	PC98
#include <pccard/pcic98reg.h>
#endif
#endif
#include <pccard/cardinfo.h>
#include <pccard/driver.h>
#include <pccard/slot.h>
#include <pccard/pcicvar.h>
/*
 *	TODO
 *		change driver.h -> pccardvar.h ??
 *				(include struct pccard_device)
 *				(include struct pccard_devinfo)
 *		change slot.h -> pccardbus.h ??
 *				(or merge to pccardvar.h)
 */

#include "pci.h"
#if NPCI > 0
#include <pci/pcivar.h>
#include <pci/pcic_p.h>
#endif
  

#define	CIRRUSHACK_FLAGS	0x1

/*
 *	Prototypes for interrupt handler.
 */
static int		pcic_ioctl __P((struct slot *, int, caddr_t));
static int		pcic_power __P((struct slot *));
static timeout_t 	pcic_reset;
static void		pcic_resume(struct slot *);
static void		pcic_disable __P((struct slot *));
static void		pcic_mapirq __P((struct slot *, int));
static timeout_t 	pcictimeout;
#ifdef LKM
static int		pcic_handle __P((struct lkm_table *lkmtp, int cmd));
#endif
static int		pcic_memory(struct slot *, int);
static int		pcic_io(struct slot *, int);
#ifdef CB_TEST
static void		pcic_status __P((struct slot *));
static int		cardbus_power __P((struct slot *));
static int		cardbus_memory(struct slot *, int);
static void		cardbus_status __P((struct slot *));
#endif /* CB_TEST */

/*
 *	Per-slot data table.
 */
static struct pcic_slot pcic_slots[MAXSLOT];

static int		pcic_irq = 0;	/*the control irq for pcic first seen*/
static unsigned		pcic_imask;
static struct slot_ctrl controller_info[NPCIC];

static int		pcic_slotnum = 0;
static int		pcic_unit = 0;

#include "isa.h"
#if NISA > 0
static int pcic_isa_probe __P((struct isa_device *));
static int pcic_isa_attach __P((struct isa_device *));
struct isa_driver pcicdriver = {
	pcic_isa_probe, pcic_isa_attach, "pcic", 0
};

static int
pcic_isa_probe(isa_dev)
	struct isa_device *isa_dev;
{
	int ports, irq, i;
	struct slot_ctrl *cinfo;

	/* isa_dev->id_unit check*/
	if (isa_dev->id_unit >= NPCIC)
		return 0;

	/* initialize controller info */
	cinfo = &controller_info[isa_dev->id_unit];

	/* used unit check*/
	if (cinfo->slot_use == 'u') {
		/* already attached */
		return 0;
	}

	/* isa_dev->id_irq is in "mask" style */
	irq = 0;
	for (i = 0; i < 16; i++) {
		if ((1 << i) & isa_dev->id_irq) {
			irq = i;
			break;
		}
	}

	ports = pcic_probe(cinfo, isa_dev->id_iobase, irq);

	return ports;
}

static int
pcic_isa_attach(dev)
	struct isa_device *dev;
{
	int irq;
	struct slot_ctrl *cinfo;

	cinfo = &controller_info[dev->id_unit];

	pcic_attach(cinfo, dev->id_flags);

	return 1;	/*XXX*/
}
#endif /* NISA */

/*
 *	Internal inline functions for accessing the PCIC.
 */
/*
 * Read a register from the PCIC.
 */
static __inline unsigned char
getb1(struct pcic_slot *sp, int reg)
{
	outb(sp->index, sp->offset + reg);
	return inb(sp->data);
}

static __inline unsigned char
getb2(struct pcic_slot *sp, int reg)
{
	return (sp->regs[reg + 0x800]);
}

static __inline unsigned char
getb(struct pcic_slot *sp, int reg)
{
	return sp->getb(sp, reg);
}

#ifdef CB_TEST
static __inline unsigned char
cb_getb(struct pcic_slot *sp, int reg)
{
	return (sp->regs[reg]);
}
  
static __inline u_long
cb_getl(struct pcic_slot *sp, int reg)
{
	u_long val = 0u;

	val |= sp->regs[reg+3] << 24;
	val |= sp->regs[reg+2] << 16;
	val |= sp->regs[reg+1] << 8;
	val |= sp->regs[reg];

	return val;
}
#endif /* CB_TEST */

/*
 * Write a register on the PCIC
 */
static __inline void
putb1(struct pcic_slot *sp, int reg, unsigned char val)
{
	outb(sp->index, sp->offset + reg);
	outb(sp->data, val);
}

static __inline void
putb2(struct pcic_slot *sp, int reg, unsigned char val)
{
	sp->regs[reg + 0x800] = val;
}

static __inline void
putb(struct pcic_slot *sp, int reg, unsigned char val)
{
	sp->putb(sp, reg, val);
}

#ifdef CB_TEST
static __inline void
cb_putb(struct pcic_slot *sp, int reg, unsigned char val)
{
	sp->regs[reg] = val;
}

static __inline void
cb_putl(struct pcic_slot *sp, int reg, u_long val)
{
	sp->regs[reg] = val & 0xFF;
	sp->regs[reg+1] = (val >> 8) & 0xFF;
	sp->regs[reg+2] = (val >> 16) & 0xFF;
	sp->regs[reg+3] = (val >> 24) & 0xFF;
}
#endif /* CB_TEST */

/*
 * Clear bit(s) of a register.
 */
static __inline void
clrb(struct pcic_slot *sp, int reg, unsigned char mask)
{
	sp->putb(sp, reg, sp->getb(sp, reg) & ~mask);
}

/*
 * Set bit(s) of a register
 */
static __inline void
setb(struct pcic_slot *sp, int reg, unsigned char mask)
{
	sp->putb(sp, reg, sp->getb(sp, reg) | mask);
}

/*
 * Write a 16 bit value to 2 adjacent PCIC registers
 */
static __inline void
putw(struct pcic_slot *sp, int reg, unsigned short word)
{
	sp->putb(sp, reg, word & 0xFF);
	sp->putb(sp, reg + 1, (word >> 8) & 0xff);
}


/*
 *	Loadable kernel module interface.
 */
#ifdef	LKM
/*
 *	This defines the lkm_misc module use by modload
 *	to define the module name.
 */
MOD_MISC(pcic);

#ifndef PCIC_INTR
#define PCIC_INTR		(1 << PCIC_IRQ)
#endif

#if NISA > 0
static struct isa_device isa_devtab_pcic[] = {
{ 0, &pcicdriver, IO_PCIC1, PCIC_INTR, -1, (caddr_t) 0, 0, pcicintr, 0, 0, 0, 0, 0, 0, 1, 0, 0 },
{ 0, &pcicdriver, IO_PCIC3, PCIC_INTR, -1, (caddr_t) 0, 0, pcicintr, 0, 0, 0, 0, 0, 0, 1, 0, 0 },
{ 0, &pcicdriver, IO_PCIC2, 0, -1, (caddr_t) 0, 0, NULL, 1, 0, 0, 0, 0, 0, 1, 0, 0 },
{ 0, &pcicdriver, IO_PCIC3, 0, -1, (caddr_t) 0, 0, NULL, 1, 0, 0, 0, 0, 0, 1, 0, 0 },
0
};
#endif

/*
 *	Module handler that processes loads and unloads.
 *	Once the module is loaded, the probe routine
 *	is called to install the slots (if any).
 */
static int
pcic_handle(struct lkm_table *lkmtp, int cmd)
{
	int err = 0;	/* default = success*/
	int found = 0;
#if NISA > 0
	struct isa_device *dvp;
#endif
#if NPCI > 0
	extern struct pci_device pcic_pci_driver;
#endif

	switch(cmd) {
	case LKM_E_LOAD:

		/*
		 * Don't load twice! (lkmexists() is exported by kern_lkm.c)
		 */
		if (lkmexists(lkmtp))
			return(EEXIST);

		bzero(pcic_slots, sizeof(pcic_slots));

		/*
		 *	Call the probe routine to find the slots. If
		 *	no slots exist, then don't bother loading the module.
		 */
#if NPCI > 0
		/* PCI PCIC probe */
		if (pci_register_lkm (&pcic_pci_driver, 0) == 0) {
#if 0
			found++;
#endif
		}
#endif
#if NISA > 0
		/* ISA PCIC probe */
		for (dvp = isa_devtab_pcic; dvp->id_driver; dvp++)
			if (pcic_isa_probe(dvp)) {
				pcic_isa_attach(dvp);
				found++;
			}
#endif

		if (!found)
			return(ENODEV);
		break;		/* Success*/
	/*
	 *	Attempt to unload the slot driver.
	 */
	case LKM_E_UNLOAD:
		printf("Unloading PCIC driver\n");
		err = pcic_unload(lkmtp, cmd);
		break;		/* Success*/

	default:	/* we only understand load/unload*/
		err = EINVAL;
		break;
	}

	return(err);
}

/*
 * External entry point; should generally match name of .o file.  The
 * arguments are always the same for all loaded modules.  The "load",
 * "unload", and "stat" functions in "DISPATCH" will be called under
 * their respective circumstances unless their value is "lkm_nullcmd".
 * If called, they are called with the same arguments (cmd is included to
 * allow the use of a single function, ver is included for version
 * matching between modules and the kernel loader for the modules).
 *
 * Since we expect to link in the kernel and add external symbols to
 * the kernel symbol name space in a future version, generally all
 * functions used in the implementation of a particular module should
 * be static unless they are expected to be seen in other modules or
 * to resolve unresolved symbols alread existing in the kernel (the
 * second case is not likely to ever occur).
 *
 * The entry point should return 0 unless it is refusing load (in which
 * case it should return an errno from errno.h).
 */
int
pcic_mod(struct lkm_table *lkmtp, int cmd, int ver)
{
	MOD_DISPATCH(pcic, lkmtp, cmd, ver,
		pcic_handle, pcic_handle, lkm_nullcmd);
}

/*
 *	pcic_unload - Called when unloading a LKM.
 *	Disables interrupts and resets PCIC.
 */
static int
pcic_unload(struct lkm_table *lkmtp, int cmd)
{
	int	slot, unit;
	struct pcic_slot *sp = pcic_slots;

	untimeout(pcictimeout, 0);
	for (slot = 0; slot < MAXSLOT; slot++, sp++) {
		if (sp->slt)
			putb(sp, PCIC_STAT_INT, 0);
		if (sp->irq)
			unregister_intr(sp->irq, pcicintr);
	}
	for (unit = 0; unit < NPCIC; unit++) {
		struct slot_ctrl *cinfo = &controller_info[unit];

		if (cinfo->slot_use == 'u')
			pccard_remove_controller(cinfo);
	}
	return(0);
}
#endif /* LKM */

#if 0
static void
pcic_dump_attributes(unsigned char *scratch, int maxlen)
{
	int i,j,k;

	i = 0;
	while (scratch[i] != 0xff && i < maxlen) {
		unsigned char link = scratch[i+2];

		/*
		 *	Dump attribute memory
		 */
		if (scratch[i]) {
			printf("[%02x] ", i);
			for (j = 0; j < 2 * link + 4 && j < 128; j += 2)
				printf("%02x ", scratch[j + i]);
			printf("\n");
		}
		i += 4 + 2 * link;
	}
}
#endif

/*
 *	entry point from main code to map/unmap memory context.
 */
static int
pcic_memory(struct slot *slt, int win)
{
	struct pcic_slot *sp = slt->cdata;
	struct mem_desc *mp = &slt->mem[win];
	int reg = mp->window * PCIC_MEMSIZE + PCIC_MEMBASE0;

#if 0
#ifdef	PC98
	if (sp->controller == PCIC_PC98) {
	    if (mp->flags & MDF_ACTIVE) {
		/* slot = 0, window = 0, sys_addr = 0xda000, length = 8KB */
		unsigned char x;
		
		if ((unsigned long)mp->start != 0xda000) {
		    printf("sys_addr must be 0xda000. requested address = 0x%x\n",
			   mp->start);
		    return(EINVAL);
		}
		
		/* omajinai ??? */
		outb(PCIC98_REG0, 0);
		x = inb(PCIC98_REG1);
		x &= 0xfc;
		x |= 0x02;
		outb(PCIC98_REG1, x);
		
		outw(PCIC98_REG_PAGOFS, 0);
		
		if (mp->flags & MDF_ATTR) {
		    outb(PCIC98_REG6, inb(PCIC98_REG6) | PCIC98_ATTRMEM);
		}else{
		    outb(PCIC98_REG6, inb(PCIC98_REG6) & (~PCIC98_ATTRMEM));
		}
		
		outb(PCIC98_REG_WINSEL, PCIC98_MAPWIN);
		
#if 0
		if (mp->flags & MDF_16BITS == 1) {	/* 16bit */
		    outb(PCIC98_REG2, inb(PCIC98_REG2) & (~PCIC98_8BIT));
		}else{					/* 8bit */
		    outb(PCIC98_REG2, inb(PCIC98_REG2) | PCIC98_8BIT);
		}
#endif
	    }else{
		outb(PCIC98_REG_WINSEL, PCIC98_UNMAPWIN);
	    }
	    return 0;
	}
#endif	/* PC98 */
#endif

	if (mp->flags & MDF_ACTIVE) {
		unsigned long sys_addr = (unsigned long)mp->start >> 12;
		/*
		 * Write the addresses, card offsets and length.
		 * The values are all stored as the upper 12 bits of the
		 * 24 bit address i.e everything is allocated as 4 Kb chunks.
		 */
		putw(sp, reg, sys_addr & 0xFFF);
		putw(sp, reg+2, (sys_addr + (mp->size >> 12) - 1) & 0xFFF);
		putw(sp, reg+4, ((mp->card >> 12) - sys_addr) & 0x3FFF);
#if 0
		printf("card offs = card_adr = 0x%x 0x%x, sys_addr = 0x%x\n", 
			mp->card, ((mp->card >> 12) - sys_addr) & 0x3FFF,
			sys_addr);
#endif
		/*
		 *	Each 16 bit register has some flags in the upper bits.
		 */
		if (mp->flags & MDF_16BITS)
			setb(sp, reg+1, PCIC_DATA16);
		if (mp->flags & MDF_ZEROWS)
			setb(sp, reg+1, PCIC_ZEROWS);
		if (mp->flags & MDF_WS0)
			setb(sp, reg+3, PCIC_MW0);
		if (mp->flags & MDF_WS1)
			setb(sp, reg+3, PCIC_MW1);
		if (mp->flags & MDF_ATTR)
			setb(sp, reg+5, PCIC_REG);
		if (mp->flags & MDF_WP)
			setb(sp, reg+5, PCIC_WP);
#if 0
	printf("Slot number %d, reg 0x%x, offs 0x%x\n",
		sp->slotnum, reg, sp->offset);
	printf("Map window to sys addr 0x%x for %d bytes, card 0x%x\n",
		mp->start, mp->size, mp->card);
	printf("regs are: 0x%02x%02x 0x%02x%02x 0x%02x%02x flags 0x%x\n",
		getb(sp, reg), getb(sp, reg+1),
		getb(sp, reg+2), getb(sp, reg+3),
		getb(sp, reg+4), getb(sp, reg+5),
		mp->flags);
#endif
		/*
		 * Enable the memory window. By experiment, we need a delay.
		 */
		setb(sp, PCIC_ADDRWINE, (1<<win) | PCIC_MEMCS16);
		DELAY(50);
	} else {
#if 0
		printf("Unmapping window %d\n", win);
#endif
		clrb(sp, PCIC_ADDRWINE, 1<<win);
		putw(sp, reg, 0);
		putw(sp, reg+2, 0);
		putw(sp, reg+4, 0);
	}
	return(0);
}

#ifdef CB_TEST
int
cardbus_memory(struct slot *slt, int win)
{
	return(0);
}
#endif /* CB_TEST */

/*
 *	pcic_io - map or unmap I/O context
 */
static int
pcic_io(struct slot *slt, int win)
{
	int	mask, reg;
	struct pcic_slot *sp = slt->cdata;
	struct io_desc *ip = &slt->io[win];
#if 0
#ifdef	PC98
	if (sp->controller == PCIC_PC98) {
	    unsigned char x;

#if 0
	    if (win =! 0) {
		printf("pcic98:Illegal PCIC I/O window request(%d)!", win);
		return(EINVAL);
	    }
#endif

	    if (ip->flags & IODF_ACTIVE) {
		unsigned short base;

		x = inb(PCIC98_REG2) & 0x0f;
		if (! (ip->flags & IODF_16BIT))
		    x |= PCIC98_8BIT;

		if (ip->size > 16)	/* 128bytes mapping */
		    x |= PCIC98_MAP128;

		x |= PCIC98_IOMEMORY;
		outb(PCIC98_REG2, x);
    
		base = 0x80d0;
		outw(PCIC98_REG4, base);		/* 98side IO base */
		outw(PCIC98_REG5, ip->start);	/* card side IO base */

#ifdef	PCIC_DEBUG
		printf("pcic98: IO mapped 0x%04x(98) -> 0x%04x(Card) and width %d bytes\n",
		       base, ip->start, ip->size);
#endif
		ip->start = base;

	    }else{
		outb(PCIC98_REG2, inb(PCIC98_REG2) & (~PCIC98_IOMEMORY));
	    }
	    return 0;
	}
#endif
#endif
	switch (win) {
	case 0:
		mask = PCIC_IO0_EN;
		reg = PCIC_IO0;
		break;
	case 1:
		mask = PCIC_IO1_EN;
		reg = PCIC_IO1;
		break;
	default:
		panic("Illegal PCIC I/O window request!");
	}
	if (ip->flags & IODF_ACTIVE) {
		unsigned char x, ioctlv;

#ifdef	PCIC_DEBUG
printf("Map I/O 0x%x (size 0x%x) on Window %d\n", ip->start, ip->size, win);
#endif	/* PCIC_DEBUG */
		putw(sp, reg, ip->start);
		putw(sp, reg+2, ip->start+ip->size-1);
		x = 0;
		if (ip->flags & IODF_ZEROWS)
			x |= PCIC_IO_0WS;
		if (ip->flags & IODF_WS)
			x |= PCIC_IO_WS;
		if (ip->flags & IODF_CS16)
			x |= PCIC_IO_CS16;
		if (ip->flags & IODF_16BIT)
			x |= PCIC_IO_16BIT;
		/*
		 * Extract the current flags and merge with new flags.
		 * Flags for window 0 in lower nybble, and in upper nybble
		 * for window 1.
		 */
		ioctlv = getb(sp, PCIC_IOCTL);
		DELAY(100);
		switch (win) {
		case 0:
			putb(sp, PCIC_IOCTL, x | (ioctlv & 0xf0));
			break;
		case 1:
			putb(sp, PCIC_IOCTL, (x << 4) | (ioctlv & 0xf));
			break;
		}
		DELAY(100);
		setb(sp, PCIC_ADDRWINE, mask);
		DELAY(100);
	} else {
		clrb(sp, PCIC_ADDRWINE, mask);
		DELAY(100);
		putw(sp, reg, 0);
		putw(sp, reg + 2, 0);
	}
	return(0);
}

#ifdef CB_TEST
extern struct cb_slot *get_cb_info(u_long iobase);
extern int get_pci_pcic_slots(u_long iobase);
#endif /* CB_TEST */
#if NPCI > 0
extern int get_pci_pcic_info(u_long iobase, pcici_t *tag);
extern char set_controller_info(pcici_t tag, struct slot_ctrl *cinfo);
#endif /* NPCI */

/*
 *	Look for an Intel PCIC (or compatible).
 *	For each available slot, allocate a PC-CARD slot.
 */
/*
 *	VLSI 82C146 has incompatibilities about the I/O address 
 *	of slot 1.  Assume it's the only PCIC whose vendor ID is 0x84,
 *	contact Nate Williams <nate@FreeBSD.org> if incorrect.
 *	rewrite: NAKAGAWA, Yoshihisa <nakagawa@jp.freebsd.org>
 */
/*
 *	TODO:
 *		separate pcic_probe() and initialize
 *		merge pcic_probe() to pcic_isa_probe()
 */
int
pcic_probe(struct slot_ctrl *cinfo, int iobase, int irq)
{
	int slotnum, i, validslots = 0;
	struct pcic_slot *sp;
	unsigned char c;
	int maybe_vlsi = 0;
	int iorange = 0;
	char oc = 0;
	int max_slots = pcic_slotnum > MAXSLOT - PCIC_MAX_SLOTS ?
				MAXSLOT - pcic_slotnum : PCIC_MAX_SLOTS;
#if NPCI > 0
	pcici_t tag;
	char pci_controller = -1;
#endif /* NPCI */
#ifdef CB_TEST
	int pci_slots;

	pci_slots = get_pci_pcic_slots(iobase);
	if (pci_slots) max_slots = pci_slots;
#endif /* CB_TEST */

	/* try to share slot irq */
	if (irq == 0 && pcic_irq) {
		printf("pcic%d: sharing irq %d with other pcic\n",
			pcic_unit, pcic_irq);
		irq = pcic_irq;
	}

	/*
	 *	Initialise controller information structure.
	 */
	cinfo->mapmem = pcic_memory;
	cinfo->mapio = pcic_io;
#ifdef CB_TEST
	cinfo->status = pcic_status;
#endif /* CB_TEST */
	cinfo->ioctl = pcic_ioctl;
	cinfo->power = pcic_power;
	cinfo->mapirq = pcic_mapirq;
	cinfo->reset = pcic_reset;
	cinfo->disable = pcic_disable;
	cinfo->resume = pcic_resume;
	cinfo->maxmem = PCIC_MEM_WIN;
	cinfo->maxio = PCIC_IO_WIN;
	cinfo->irqs = 0;		/* set pccard.c */
	cinfo->imask = &pcic_imask;	/* XXXX pccard.c ? */
	cinfo->slots = 0;
	cinfo->name = "Unknown!";
#ifdef CB_TEST
	cinfo->cardbus = 0;
#endif /* CB_TEST */
	cinfo->next = NULL;

id_recheck:
	sp = &pcic_slots[pcic_slotnum];
	for (slotnum = 0; slotnum < max_slots; slotnum++, sp++) {
		/*
		 *	Initialise the PCIC slot table.
		 */
#if NPCI > 0
		if (get_pci_pcic_info(iobase, &tag)) {
			/* override controller info */
			pci_controller = set_controller_info(tag, cinfo);
		}
#endif /* NPCI */
#ifdef CB_TEST
		if (!cinfo->cardbus) {
#endif /* CB_TEST */
			sp->getb = getb1;
			sp->putb = putb1;
			sp->index = iobase;
			sp->data = iobase + 1;
			sp->regs = 0;
#ifdef CB_TEST
		} else {
			struct cb_slot *cb_info;

			cb_info = get_cb_info(iobase);
			if (cb_info) {
				cb_info += slotnum;
				sp->regs = (u_char *)cb_info->cb_base;
			}
			sp->getb = getb2;
			sp->putb = putb2;
			sp->index = 0;
			sp->data = 0;
#if 0
			sp->regs = (u_char *)iobase;
			max_slots = 1;
#endif
		}
#endif /* CB_TEST */
		iorange = iorange < 2 ? 2 : iorange;
		sp->offset = slotnum * PCIC_SLOT_SIZE;
		sp->irq = irq;

#ifdef CB_TEST
		if (pci_controller >= 0) {
			sp->controller = pci_controller;
		} else {
#endif /* CB_TEST */
		/* 
		 * XXX - Screwed up slot 1 on the VLSI chips.  According to
		 * the Linux PCMCIA code from David Hinds, working chipsets
		 * return 0x84 from their (correct) ID ports, while the broken
		 * ones would need to be probed at the new offset we set after
		 * we assume it's broken.
		 */
		if (maybe_vlsi && slotnum == 1) {
			sp->index += 4;
			sp->data += 4;
			sp->offset = PCIC_SLOT_SIZE << 1;
			iorange = iorange < 6 ? 6 : iorange;
		}
		/*
		 * see if there's a PCMCIA controller here
		 * Intel PCMCIA controllers use 0x82 and 0x83
		 * IBM clone chips use 0x88 and 0x89, apparently
		 */
		c = getb(sp, PCIC_ID_REV);
		if (bootverbose)
			printf("PCIC ID = 0x%02x\n", (u_int)c);
		sp->revision = -1;
		switch(c) {
		/*
		 *	82365 or clones.
		 */
		case 0x82:
		case 0x83:
			sp->controller = PCIC_I82365;
			sp->revision = c & 1;
			/*
			 *	Now check for VADEM chips.
			 */
			outb(sp->index, 0x0E);
			outb(sp->index, 0x37);
			setb(sp, 0x3A, 0x40);
			c = getb(sp, PCIC_ID_REV);
			if (c & 0x08) {
				sp->controller = ((sp->revision = c & 7) == 4) ?	
					PCIC_VG469 : PCIC_VG468 ;
				clrb(sp, 0x3A, 0x40);
			}

			/*
			 * Check for RICOH RF5C396 PCMCIA Controller
			 */
			c = getb(sp, 0x3a);
			if (c == 0xb2) {
				sp->controller = PCIC_RF5C396;
			}

			break;
		/*
		 *	82365DF or clones or VLSI.
		 */
		case 0x84:
			if (!maybe_vlsi) {
				sp->controller = PCIC_I82365DF;
			} else {
				sp->controller = PCIC_VLSI;
			}
			break;
		case 0x88:
		case 0x89:
			sp->controller = PCIC_IBM;
			sp->revision = c & 1;
			break;
		case 0x8a:
			sp->controller = PCIC_IBM_KING;
			break;
		default:
			/*
			 *	check VLSI chips.
			 */
			if (slotnum == 1 && oc == PCIC_I82365DF) {
				maybe_vlsi = 1;
				validslots = 0;
				goto id_recheck;
			}
			continue;
		}
		/* store old controller type */
		oc = sp->controller;

		/*
		 *	Check for Cirrus logic chips.
		 */
		putb(sp, 0x1F, 0);
		c = getb(sp, 0x1F);
		if ((c & 0xC0) == 0xC0) {
			c = getb(sp, 0x1F);
			if ((c & 0xC0) == 0) {
				if (c & 0x20)
					sp->controller = PCIC_PD672X;
				else
					sp->controller = PCIC_PD6710;
				sp->revision = 8 - ((c & 0x1F) >> 2);
			}
		}
#if NPCI > 0
		if (pci_controller >= 0) {
			sp->controller = pci_controller;
		} else {
#endif /* NPCI */
		switch(sp->controller) {
		case PCIC_I82365:
			cinfo->name = "Intel 82365A/B";
			break;
		case PCIC_I82365DF:
			cinfo->name = "Intel 82365DF";
			break;
		case PCIC_IBM:
			cinfo->name = "IBM PCIC";
			break;
		case PCIC_IBM_KING:
			cinfo->name = "IBM KING PCMCIA Controller";
			break;
		case PCIC_PD672X:
			cinfo->name = "Cirrus Logic PD672X";
			break;
		case PCIC_PD6710:
			cinfo->name = "Cirrus Logic PD6710";
			break;
		case PCIC_VG468:
			cinfo->name = "Vadem 468";
			break;
		case PCIC_VG469:
			cinfo->name = "Vadem 469";
			break;
		case PCIC_RF5C396:
			cinfo->name = "Ricoh RF5C396";
			break;
		case PCIC_VLSI:
			cinfo->name = "VLSI 82C146";
			break;
		default:
			cinfo->name = "Unknown!";
			break;
		}
#if NPCI > 0
		}
#endif /* NPCI */
#ifdef CB_TEST
		}
#endif /* CB_TEST */
		validslots++;
	}
	cinfo->slots = validslots;

	return validslots ? iorange : 0;
}

/*
 *	This code is for a PC which the ExCA timing register isn't
 *	defined by BIOS. When you use this, it is necessary to write
 *	"flags 0x1" in a line of pcic? in kernel configuration file. 
 *	Takeshi Shibagaki(shiba@jp.freebsd.org)
 */
static void
cirrus_hack(struct pcic_slot *sp)
{
	switch (sp->controller) {
		case PCIC_PD6710:
		case PCIC_PD672X:
			sp->putb(sp, PCIC_TIME_SETUP0, 0x01);
			sp->putb(sp, PCIC_TIME_CMD0,   0x06);
			sp->putb(sp, PCIC_TIME_RECOV0, 0x03);
			sp->putb(sp, PCIC_TIME_SETUP1, 0x01);
			sp->putb(sp, PCIC_TIME_CMD1,   0x0f);
			sp->putb(sp, PCIC_TIME_RECOV1, 0x03);
			break;
		case PCIC_PD6729:
			sp->putb(sp, PCIC_TIME_SETUP0, 0x01);
			sp->putb(sp, PCIC_TIME_CMD0,   0x05);
			sp->putb(sp, PCIC_TIME_RECOV0, 0x00);
			sp->putb(sp, PCIC_TIME_SETUP1, 0x01);
			sp->putb(sp, PCIC_TIME_CMD1,   0x14);
			sp->putb(sp, PCIC_TIME_RECOV1, 0x00);
			break;
		case PCIC_PD6832:
			sp->putb(sp, PCIC_TIME_SETUP0, 0x00);
			sp->putb(sp, PCIC_TIME_CMD0,   0x07);
			sp->putb(sp, PCIC_TIME_RECOV0, 0x04);
			sp->putb(sp, PCIC_TIME_SETUP1, 0x03);
			sp->putb(sp, PCIC_TIME_CMD1,   0x1d);
			sp->putb(sp, PCIC_TIME_RECOV1, 0x04);
			break;
  	}
	/*
	 *	This code is to turn on speaker enable bit in
	 *	the ExCA register.
	 *	In fact, the bit must be defined by BIOS.
	 *	I don't know a reason that the Low-Power Dynamic Mode
	 *	enable bit turn on to '1'. 
	 *	Please justify, someone.
	 *	Takeshi Shibagaki(shiba@jp.freebsd.org)
	 */
	switch (sp->controller) {
		case PCIC_PD6710:
		case PCIC_PD672X:
		case PCIC_PD6729:
			setb(sp, CL672X_PCIC_MISC2, CL672X_M2_LPDM_EN);
		case PCIC_PD6832:
			setb(sp, CL672X_PCIC_MISC1, CL672X_M1_SPKR_EN);
			break;
	}
}

int
pcic_attach(struct slot_ctrl *cinfo, int flags)
{
	int slotnum, i;
	struct slot *slt;
	struct pcic_slot *sp;
#ifdef CB_TEST
	struct cb_slot *cb_info;
#endif /* CB_TEST */

	pccard_add_controller(cinfo);
	sp = &pcic_slots[pcic_slotnum];
	for (slotnum = 0; slotnum < cinfo->slots; slotnum++, sp++) {
		/*
		 *	OK it seems we have a PCIC or lookalike.
		 *	Allocate a slot and initialise the data structures.
		 */
		sp->slotnum = slotnum;
		slt = pccard_alloc_slot(cinfo);
		if (slt == 0)
			continue;
		printf("pcic%d: slot %d controller I/O address 0x%x\n",
			pcic_unit, slotnum, sp->index);
		slt->cdata = sp;
		sp->slt = slt;
		/*
		 *	If we haven't allocated an interrupt for the controller,
		 *	then attempt to get one.
		 */
		if (pcic_irq == 0) {
			pcic_imask = SWI_MASK;
			if (sp->irq)
				pcic_irq = sp->irq;
			else {
				printf(
"pcic%d: no irq specified, no previous pcic irq found\n", pcic_unit);
			}
		}
		/*
		 *	Check for a card in this slot.
		 */
		setb(sp, PCIC_POWER, PCIC_PCPWRE| PCIC_DISRST);
		if ((getb(sp, PCIC_STATUS) & PCIC_CD) != PCIC_CD) {
			slt->laststate = slt->state = empty;
		} else {
			slt->laststate = slt->state = filled;
			pccard_event(sp->slt, card_inserted);
		}
		/*
		 *	Assign IRQ for slot changes
		 */
		if (sp->irq > 0)
			putb(sp, PCIC_STAT_INT, (sp->irq << 4) | 0xF);
		/*
		 *	If flags 0x1 is defined in kernel configuration file,
		 *	jump to cirrus pcic specific code.
		 */
		if((flags & CIRRUSHACK_FLAGS) == CIRRUSHACK_FLAGS)
			cirrus_hack(sp);
	}
#if 0
#ifdef	PC98
	if (validslots == 0) {
	    sp = pcic_slots;
	    slotnum = 0;
	    if (inb(PCIC98_REG0) != 0xff) {
		sp->controller = PCIC_PC98;
		sp->revision = 0;
		cinfo->name = "PC98 Original";
		cinfo->maxmem = 1;
		cinfo->maxio = 1;
/*		cinfo->irqs = PCIC_INT_MASK_ALLOWED;*/
		cinfo->irqs = 0x1468;
		validslots++;
		sp->slotnum = slotnum;

		slt = pccard_alloc_slot(cinfo);
		if (slt == 0) {
		    printf("pcic98: slt == NULL\n");
		    goto pcic98_probe_end;
		}
		slt->cdata = sp;
		sp->slt = slt;

		/* Check for a card in this slot */
		if (inb(PCIC98_REG1) & PCIC98_CARDEXIST) {
		    /* PCMCIA card exist */
		    slt->laststate = slt->state = filled;
		    pccard_event(sp->slt, card_inserted);
		} else {
		    slt->laststate = slt->state = empty;
		}
	    }
	pcic98_probe_end:
	}
#endif	/* PC98 */
#endif
	if (cinfo->slots) {
		timeout(pcictimeout, 0, hz/2);
		cinfo->slot_use = 'u';
	}

	pcic_slotnum += cinfo->slots;
	pcic_unit++;

	return 1;
}

/*
 *	ioctl calls - Controller specific ioctls
 */
static int
pcic_ioctl(struct slot *slt, int cmd, caddr_t data)
{
	struct pcic_slot *sp = slt->cdata;

	switch(cmd) {
	default:
		return(EINVAL);
	/*
	 * Get/set PCIC registers
	 */
	case PIOCGREG:
		((struct pcic_reg *)data)->value =
			getb(sp, ((struct pcic_reg *)data)->reg);
		break;
	case PIOCSREG:
		putb(sp, ((struct pcic_reg *)data)->reg,
			((struct pcic_reg *)data)->value);
		break;
	}
	return(0);
}

#ifdef CB_TEST
/*
 *	pcic_status - Get slot status
 */
static void
pcic_status(struct slot *slt)
{
	struct pcic_slot *sp = slt->cdata;
	unsigned char reg;
	int stschg = 0, detect = 0, power = 0;
	int ready = 0;

	reg = getb(sp, PCIC_STATUS);
	if ((reg & PCIC_CD) == PCIC_CD)
		detect = 1;
	if (reg & PCIC_POW)
		power = 1;
	if (reg & PCIC_READY)
		ready = 1;

	if (sp->controller == PCIC_PD672X) {
		reg = getb (sp, CL672X_PCIC_MISC1);
		if (!(reg & CL672X_M1_5V_DET))
			slt->pwr.vcc = 33;
	}
	if (sp->controller == PCIC_VG469) {
		reg = getb (sp, VG469_VSENSE);
		if (slt->ctrl->slots & 1) {
			if (!(reg & VG469_VSENSE_B_VS1))
				slt->pwr.vcc = 33;
		} else {
			if (!(reg & VG469_VSENSE_A_VS1))
				slt->pwr.vcc = 33;
		}
	}

	return;
}

void
cardbus_status(struct slot *slt)
{
	struct pcic_slot *sp = slt->cdata;
	u_long reg;
	int stschg = 0, detect = 0, power = 0, cardbus = 0;
	int ready = 0;

	if (slt->ctrl->cardbus) {
		reg = cb_getl(sp, CB_PRESENT_STATE);
		if (reg & CB_CSTSCHG)
			stschg = 1;
		if (!(reg & CB_CCD1) && !(reg & CB_CCD2))
			detect = 1;
		if (reg & CB_POWER_CYCLE) {
			power = 1;
			ready = 1;
		}
		if (reg & CB_CB_CARD)
			cardbus = 1;
		if (reg & CB_5V_CARD)
			slt->pwr.vcc = 50;
		if (reg & CB_3V_CARD)
			slt->pwr.vcc = 33;
	}

	return;
}
#endif /* CB_TEST */

/*
 *	pcic_power - Enable the power of the slot according to
 *	the parameters in the power structure(s).
 */
static int
pcic_power(struct slot *slt)
{
	unsigned char reg = PCIC_DISRST|PCIC_PCPWRE;
	struct pcic_slot *sp = slt->cdata;

	switch(sp->controller) {
#if 0
#ifdef	PC98
	case PCIC_PC98:
	    reg = inb(PCIC98_REG6) & (~PCIC98_VPP12V);
	    switch(slt->pwr.vpp) {
	    default:
		return(EINVAL);
	    case 50:
		break;
	    case 120:
		reg |= PCIC98_VPP12V;
		break;
	    }
	    outb(PCIC98_REG6, reg);
	    DELAY(100*1000);

	    reg = inb(PCIC98_REG2) & (~PCIC98_VCC3P3V);
	    switch(slt->pwr.vcc) {
	    default:
		return(EINVAL);
	    case 33:
		reg |= PCIC98_VCC3P3V;
		break;
	    case 50:
		break;
	    }
	    outb(PCIC98_REG2, reg);
	    DELAY(100*1000);
	    return (0);
#endif
#endif
	case PCIC_PD6832:
	case PCIC_PD6729:
	case PCIC_PD672X:
	case PCIC_PD6710:
	case PCIC_VG468:
	case PCIC_VG469:
	case PCIC_RF5C396:
	case PCIC_VLSI:
	case PCIC_IBM_KING:
	case PCIC_I82365DF:
	case PCIC_TI1130:
	case PCIC_TI1131:
	case PCIC_TI1220:
	case PCIC_TI1221:
	case PCIC_TI1250:
		switch(slt->pwr.vpp) {
		default:
			return(EINVAL);
		case 0:
			break;
		case 50:
		case 33:
			reg |= PCIC_VPP_5V;
			break;
		case 120:
			reg |= PCIC_VPP_12V;
			break;
		}
		switch(slt->pwr.vcc) {
		default:
			return(EINVAL);
		case 0:
			break;
		case 33:
			if (sp->controller == PCIC_IBM_KING) {
				reg |= PCIC_VCC_5V_KING;
				break;
			}
			reg |= PCIC_VCC_3V;
			if ((sp->controller == PCIC_VG468)||
				(sp->controller == PCIC_VG469))
				setb(sp, 0x2f, 0x03) ;
			else
				setb(sp, PCIC_CDGC, 0x02);
			break;
		case 50:
			if (sp->controller == PCIC_IBM_KING) {
				reg |= PCIC_VCC_5V_KING;
				break;
			}
			reg |= PCIC_VCC_5V;
			if ((sp->controller == PCIC_VG468)||
				(sp->controller == PCIC_VG469))
				clrb(sp, 0x2f, 0x03) ;
			else
				clrb(sp, PCIC_CDGC, 0x02);
			break;
		}
		break;
	}
	putb(sp, PCIC_POWER, reg);
	DELAY(300*1000);
	if (slt->pwr.vcc) {
		reg |= PCIC_OUTENA;
		putb(sp, PCIC_POWER, reg);
		DELAY(100*1000);
	}
	/* Some chips are smarter than us it seems, so if we weren't
	 * allowed to use 5V, try 3.3 instead
	 */
	if (!(getb(sp, PCIC_STATUS) &  0x40) && slt->pwr.vcc == 50) {
		slt->pwr.vcc = 33;
		slt->pwr.vpp = 0;
		return (pcic_power(slt));
	}
	return(0);
}

#ifdef CB_TEST
int
cardbus_power(struct slot *slt)
{
	u_long reg = 0u;
	struct pcic_slot *sp = slt->cdata;
	struct slot_ctrl *cinfo = slt->ctrl;
	u_int16_t brgcntl;

	if (slt->ctrl->cardbus) {
		switch(slt->pwr.vpp) {
		default:
			return(EINVAL);
		case 0:
			break;
		case 50:
			reg |= CB_VPP_5V;
			break;
		case 33:
			reg |= CB_VPP_3V;
			break;
		case 120:
			reg |= CB_VPP_12V;
			break;
		}
		switch(slt->pwr.vcc) {
		default:
			return(EINVAL);
		case 0:
			break;
		case 50:
			reg |= CB_VCC_5V;
			break;
		case 33:
			reg |= CB_VCC_3V;
			break;
		}
		if (reg != cb_getl(sp, CB_CONTROL)) {
			cb_putl(sp, CB_CONTROL, reg);
			DELAY(100*1000);
		}

		/* Bridge Control */
#if 0 /* XXXX */
		brgcntl = pci_cfgread(cinfo->tag, CB_PCI_BRIDGE_CTRL, 2);
		brgcntl &= ~CB_BCR_CB_RESET;
		pci_cfgwrite(cinfo->tag, CB_PCI_BRIDGE_CTRL, brgcntl, 2);
#endif

#if 1
		reg = cb_getl(sp, CB_PRESENT_STATE);
		if (reg & CB_POWER_CYCLE)
			printf("powered up\n");
		else
			printf("not yet powered up\n");
#endif
	}
	return(0);
}
#endif /* CB_TEST */

/*
 * tell the PCIC which irq we want to use.  only the following are legal:
 * 3, 4, 5, 7, 9, 10, 11, 12, 14, 15
 */
static void
pcic_mapirq(struct slot *slt, int irq)
{
	struct pcic_slot *sp = slt->cdata;
	u_char x;

#if 0
#ifdef	PC98
	if (sp->controller == PCIC_PC98) {
	    switch (irq) {
	    case 3:
		x = PCIC98_INT0; break;
	    case 5:
		x = PCIC98_INT1; break;
	    case 6:
		x = PCIC98_INT2; break;
	    case 10:
		x = PCIC98_INT4; break;
	    case 12:
		x = PCIC98_INT5; break;
	    case 0:		/* disable */
		x = PCIC98_INTDISABLE;
		break;
	    default:
		printf("pcic98: illegal irq %d\n", irq);
		return;
	    }
#ifdef	PCIC_DEBUG
	    printf("pcic98: irq=%d mapped.\n", irq);
#endif
	    outb(PCIC98_REG3, x);

	    return;
	}	
#endif
#endif
	if (irq == 0)
		clrb(sp, PCIC_INT_GEN, 0xF);
	else
		putb(sp, PCIC_INT_GEN, 
		    (getb(sp, PCIC_INT_GEN) & 0xF0) | irq);
#if NPCI > 0
	switch (sp->controller) {
	/* CLPD-6832 ISA-mode interrupt support (experimental) */
	case PCIC_PD6832:
		/* Read "Misc Control 3" register of PD-6832 */
		putb(sp, CL6832_ExCA_EXTEND_INDEX, CL6832_ExCA_EXT_MISC3);
		x = getb(sp, CL6832_ExCA_EXTEND_DATA);
#if 1
printf("Experimental clpd-6832 support:\n");
printf("please send the following messages to hosokawa@jp.FreeBSD.org\n");
printf("Misc 3 register is 0x%x\n", (u_int)x);
#endif
		x = ((x & ~0x3) | 0x1);	/* set "external hardware" bit */
#if 1
printf("Set Misc 3 register to 0x%x\n", (u_int)x);
#endif
		putb(sp, CL6832_ExCA_EXTEND_INDEX, CL6832_ExCA_EXT_MISC3);
		putb(sp, CL6832_ExCA_EXTEND_DATA, x);

		/* check it again */
		putb(sp, CL6832_ExCA_EXTEND_INDEX, CL6832_ExCA_EXT_MISC3);
		x = getb(sp, CL6832_ExCA_EXTEND_DATA);
#if 1
printf("Misc 3 register is 0x%x\n", (u_int)x);
#endif
	}
#endif	/* NPCI > 0 */
}

#ifdef CB_TEST
static void
cardbus_print_register(struct slot *slt)
{
	struct pcic_slot *sp = (struct pcic_slot *)slt->cdata;

	if (slt->ctrl->cardbus) {
		int i;

		printf ("Cardbus Socket registers:\n");
		printf("00: ");
		for (i=0;i<16;i+=4)
			printf(" %08x:", cb_getl(sp, i));
		printf("\n10: ");
		for (i=16;i<32;i+=4)
			printf(" %08x:", cb_getl(sp, i));
		printf("\n");
	}
}
#endif /* CB_TEST */

/*
 *	pcic_reset - Reset the card and enable initial power.
 */
static void
pcic_reset(void *chan)
{
	struct slot *slt = chan;
	struct pcic_slot *sp = slt->cdata;

#ifdef CB_TEST
	if (bootverbose) {
		printf("pcic_reset\n");
		cardbus_print_register(slt);
	}
#endif /* CB_TEST */

#if 0
#ifdef	PC98
	if (sp->controller == PCIC_PC98) {
	    outb(PCIC98_REG0, 0);
	    outb(PCIC98_REG2, inb(PCIC98_REG2) & (~PCIC98_IOMEMORY));
	    outb(PCIC98_REG3, PCIC98_INTDISABLE);
	    outb(PCIC98_REG2, inb(PCIC98_REG2) & (~PCIC98_VCC3P3V));
	    outb(PCIC98_REG6, inb(PCIC98_REG6) & (~PCIC98_VPP12V));
	    outb(PCIC98_REG1, 0);

	    selwakeup(&slt->selp);
	    return;
	}
#endif
#endif
	switch (slt->insert_seq) {
	    case 0: /* Something funny happended on the way to the pub... */
		return;
	    case 1: /* Assert reset */
		clrb(sp, PCIC_INT_GEN, PCIC_CARDRESET);
		slt->insert_seq = 2;
		timeout(pcic_reset, (void *)slt, hz/4);
		return;
	    case 2: /* Deassert it again */
		setb(sp, PCIC_INT_GEN, PCIC_CARDRESET|PCIC_IOCARD);
		slt->insert_seq = 3;
		timeout(pcic_reset, (void *)slt, hz/4);
		return;
	    case 3: /* Wait if card needs more time */
		if (!getb(sp, PCIC_STATUS) & PCIC_READY) {
			timeout(pcic_reset, (void *)slt, hz/10);
			return;
		}
	}
	slt->insert_seq = 0;
	selwakeup(&slt->selp);
}

/*
 *	pcic_disable - Disable the slot.
 */
static void
pcic_disable(struct slot *slt)
{
	struct pcic_slot *sp = slt->cdata;

#if 0
#ifdef	PC98
	if (sp->controller == PCIC_PC98) {
	    return;
	}
#endif
#endif
	putb(sp, PCIC_INT_GEN, 0);
	putb(sp, PCIC_POWER, 0);
}

/*
 *	PCIC timer, it seems that we lose interrupts sometimes
 *	so poll just in case...
 */
static void
pcictimeout(void *chan)
{
	pcicintr(0);
	timeout(pcictimeout, 0, hz/2);
}

/*
 *	PCIC Interrupt handler.
 *	Check each slot in turn, and read the card status change
 *	register. If this is non-zero, then a change has occurred
 *	on this card, so send an event to the main code.
 */
void
pcicintr(int unit)
{
	int	slot, s;
	unsigned char chg;
	struct pcic_slot *sp = pcic_slots;

#if 0
#ifdef	PC98
	if (sp->controller == PCIC_PC98) {
	    slot = 0;
	    s = splhigh();
	    /* Check for a card in this slot */
	    if (inb(PCIC98_REG1) & PCIC98_CARDEXIST) {
		if (sp->slt->laststate != filled) {
		    pccard_event(sp->slt, card_inserted);
		}
	    } else {
		if (sp->slt->laststate != empty) {
		    pccard_event(sp->slt, card_removed);
		}
	    }
	    splx(s);
	    return;
	}
#endif	/* PC98 */
#endif
	s = splhigh();
	for (slot = 0; slot < PCIC_MAX_SLOTS * NPCIC; slot++, sp++) {
		if (sp->slt && (chg = getb(sp, PCIC_STAT_CHG)) != 0)
			if (chg & PCIC_CDTCH) {
#ifdef CB_TEST
				if (bootverbose) {
					printf("pcicintr\n");
					cardbus_print_register(sp->slt);
				}
				if (sp->slt->ctrl->cardbus) {
					u_long reg;

					reg = cb_getl(sp, CB_PRESENT_STATE);
					if (bootverbose) {
						if (reg & CB_16BIT_CARD)
							printf("16bit card\n");
						if (reg & CB_CB_CARD)
							printf("CardBus card\n");
						if (reg & CB_5V_CARD)
							printf("5V card\n");
						if (reg & CB_3V_CARD)
							printf("3V card\n");
					}
				}
#endif /* CB_TEST */
				if ((getb(sp, PCIC_STATUS) & PCIC_CD) ==
						PCIC_CD) {
					pccard_event(sp->slt, card_inserted);
#ifdef CB_TEST
					if (bootverbose) {
						printf("inserted\n");
						cardbus_print_register(sp->slt);
					}
#endif /* CB_TEST */
				} else {
					pccard_event(sp->slt, card_removed);
				}
			}
	}
	splx(s);
}

#ifdef CB_TEST
void
pcic_intr_sc(void *arg)
{
	int	slot, s;
	unsigned char chg;
	struct slot_ctrl *cinfo = (struct slot_ctrl *)arg;
	struct slot *slt;
	struct pcic_slot *sp;

#if 0
	/* intr slot_list per controller */
	for (slt = cinfo->slot_list; slt; slt = slt->next) {
		sp = (struct pcic_slot *)slt->cdata;
		pcic_status(slt);
		if (status)
			pccard_event(slt, card_inserted);
		else
			pccard_event(slt, card_removed);
	}
#else
	pcicintr(0); /* XXXX */
#endif
}
#endif /* CB_TEST */

/*
 *	pcic_resume - Suspend/resume support for PCIC
 */
static void
pcic_resume(struct slot *slt)
{
	struct pcic_slot *sp = slt->cdata;
	if (sp->irq > 0)
		putb(sp, PCIC_STAT_INT, (sp->irq << 4) | 0xF);
}
#endif  /* NPCIC */
